<?php declare(strict_types=1);
/*
 * This file is part of sebastian/diff.
 *
 * (c) Sebastian Bergmann <sebastian@phpunit.de>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */
namespace SebastianBergmann\Diff;

use LogicException;

/**
 * @codeCoverageIgnore
 */
final class NoMiddleSnakeFoundException extends LogicException implements Exception
{
    public function __construct()
    {
        parent::__construct('No middle snake found; input invariants violated');
    }
}
